@@ -1,5 +1,7 @@
-/* Copyright 2006 Fluendo S.L. 
- * Author: Jan Schmidt <jan@fluendo.com>
+/* Copyright 2006, 2007, 2008 Fluendo S.A. 
+ * Authors: Jan Schmidt <jan@fluendo.com>
+ *          Kapil Agrawal <kapil@fluendo.com>
+ *          Julien Moutte <julien@fluendo.com>
  * See the COPYING file in the top-level directory.
  */
 
@@ -10,9 +12,17 @@
 
 #include "flumpegtsmux.h"
 
+#include "flumpegtsmux_h264.h"
+#include "flumpegtsmux_aac.h"
+
 GST_DEBUG_CATEGORY (flutsmux_debug);
 #define GST_CAT_DEFAULT flutsmux_debug
 
+enum{
+  ARG_0,
+  ARG_M2TS_MODE
+};
+
 static GstStaticPadTemplate flutsmux_sink_factory =
 GST_STATIC_PAD_TEMPLATE ("sink_%d",
     GST_PAD_SINK,
@@ -20,7 +30,8 @@ GST_STATIC_PAD_TEMPLATE ("sink_%d",
     GST_STATIC_CAPS ("video/mpeg, mpegversion=(int) { 1, 2 }, "
                      "systemstream = (boolean) false; "
                      "video/x-dirac;"
-                     "audio/mpeg, mpegversion = (int) { 1, 2 }"
+                     "video/x-h264;"
+                     "audio/mpeg, mpegversion = (int) { 1, 2, 4 }"
     )
 );
 
@@ -28,11 +39,20 @@ static GstStaticPadTemplate flutsmux_src_factory =
 GST_STATIC_PAD_TEMPLATE ("src",
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
-    GST_STATIC_CAPS ("video/mpegts")
+    GST_STATIC_CAPS ("video/mpegts, "
+                     "systemstream = (boolean) true, "
+                     "packetsize = (int) { 188, 192} "
+                    )
 );
 
+static void gst_flutsmux_set_property (GObject * object, guint prop_id,
+    const GValue * value, GParamSpec * pspec);
+static void gst_flutsmux_get_property (GObject * object, guint prop_id,
+    GValue * value, GParamSpec * pspec);
+
 static void flutsmux_dispose (GObject * object);
-static gboolean new_packet_cb (guint8 *data, guint len, void *user_data);
+static gboolean new_packet_cb (guint8 *data, guint len, void *user_data, 
+    gint64 new_pcr);
 static void release_buffer_cb (guint8 *data, void *user_data);
 
 static gboolean flutsdemux_prepare_srcpad (FluTsMux *mux);
@@ -71,11 +91,20 @@ flutsmux_class_init (FluTsMuxClass *klass)
   GstElementClass *gstelement_class = GST_ELEMENT_CLASS (klass);
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
 
+  gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_flutsmux_set_property);
+  gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_flutsmux_get_property);
   gobject_class->dispose = flutsmux_dispose;
 
   gstelement_class->request_new_pad = flutsmux_request_new_pad;
   gstelement_class->release_pad = flutsmux_release_pad;
   gstelement_class->change_state = flutsmux_change_state;
+
+   g_object_class_install_property (G_OBJECT_CLASS (klass), ARG_M2TS_MODE,
+        g_param_spec_boolean ("m2ts_mode", "M2TS(192 bytes) Mode",
+        "Defines what packet size to use, normal TS format ie .ts(188 bytes) "
+        "or Blue-Ray disc ie .m2ts(192 bytes).",
+        FALSE, G_PARAM_READWRITE));
+
 }
 
 static void
@@ -97,13 +126,22 @@ flutsmux_init (FluTsMux *mux, FluTsMuxClass *g_class)
 
   mux->first = TRUE;
   mux->last_flow_ret = GST_FLOW_OK;
+  mux->adapter = gst_adapter_new();
+  mux->m2ts_mode = FALSE;
+  mux->first_pcr = TRUE;
+  mux->last_ts = 0;
 }
 
 static void
 flutsmux_dispose (GObject * object)
 {
   FluTsMux *mux = FLU_TSMUX (object);
-  
+ 
+  if(mux->adapter){
+    gst_adapter_clear(mux->adapter);
+    gst_object_unref(mux->adapter);
+    mux->adapter = NULL; 
+  }  
   if (mux->collect) {
     gst_object_unref (mux->collect);
     mux->collect = NULL;
@@ -116,6 +154,39 @@ flutsmux_dispose (GObject * object)
   GST_CALL_PARENT (G_OBJECT_CLASS, dispose, (object));
 }
 
+static void
+gst_flutsmux_set_property (GObject * object, guint prop_id,
+      const GValue * value, GParamSpec * pspec)
+{
+  FluTsMux * mux = FLU_TSMUX (object);
+
+  switch (prop_id) {
+    case ARG_M2TS_MODE:
+      /*set incase if the output stream need to be of 192 bytes*/      
+      mux->m2ts_mode = g_value_get_boolean (value);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+}
+
+static void
+gst_flutsmux_get_property (GObject * object, guint prop_id,
+      GValue * value, GParamSpec * pspec)
+{
+  FluTsMux * mux = FLU_TSMUX (object);
+
+  switch (prop_id) {
+    case ARG_M2TS_MODE:
+      g_value_set_boolean (value, mux->m2ts_mode);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+}
+
 static void release_buffer_cb (guint8 *data, void *user_data)
 {
   GstBuffer *buf = (GstBuffer *) user_data;
@@ -130,7 +201,7 @@ static GstFlowReturn flutsmux_create_stream (FluTsMux *mux,
   GstStructure *s;
 
   if (caps == NULL) {
-    GST_DEBUG_OBJECT (mux, "Sink pad caps were not set before pushing");
+    GST_DEBUG_OBJECT (pad, "Sink pad caps were not set before pushing");
     return GST_FLOW_NOT_NEGOTIATED;
   }
   
@@ -138,33 +209,74 @@ static GstFlowReturn flutsmux_create_stream (FluTsMux *mux,
   g_return_val_if_fail (s != NULL, FALSE);
 
   if (gst_structure_has_name (s, "video/x-dirac")) {
-    GST_DEBUG_OBJECT (mux, "Creating Dirac stream with PID 0x%04x", 
+    GST_DEBUG_OBJECT (pad, "Creating Dirac stream with PID 0x%04x", 
         ts_data->pid);
     ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_VIDEO_DIRAC, 
         ts_data->pid);
   }
+  else if (gst_structure_has_name (s, "video/x-h264")) {
+    const GValue * value;
+    GST_DEBUG_OBJECT (pad, "Creating H264 stream with PID 0x%04x", 
+        ts_data->pid);
+    /* Codec data contains SPS/PPS which need to go in stream for valid ES */
+    value = gst_structure_get_value (s, "codec_data");
+    if (value) {
+      ts_data->codec_data = gst_buffer_ref (gst_value_get_buffer (value));
+      GST_DEBUG_OBJECT (pad, "we have additional codec data (%d bytes)",
+          GST_BUFFER_SIZE (ts_data->codec_data));
+      ts_data->prepare_func = flutsmux_prepare_h264;
+    }
+    else {
+      ts_data->codec_data = NULL;
+    }
+    ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_VIDEO_H264, 
+        ts_data->pid);
+  }
   else if (gst_structure_has_name (s, "audio/mpeg")) {
     gint mpegversion;
     if (!gst_structure_get_int (s, "mpegversion", &mpegversion)) {
-      GST_ELEMENT_ERROR (mux, STREAM, FORMAT, 
+      GST_ELEMENT_ERROR (pad, STREAM, FORMAT, 
           ("Invalid data format presented"), 
           ("Caps with type audio/mpeg did not have mpegversion"));
       goto beach;
     }
 
-    if (mpegversion == 1) {
-      GST_DEBUG_OBJECT (mux, 
-          "Creating MPEG Audio, version 1 stream with PID 0x%04x", 
-          ts_data->pid);
-      ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_AUDIO_MPEG1, 
-        ts_data->pid);
-    }
-    else {
-      GST_DEBUG_OBJECT (mux, 
-          "Creating MPEG Audio, version 2 stream with PID 0x%04x", 
-          ts_data->pid);
-      ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_AUDIO_MPEG2, 
-        ts_data->pid);
+    switch (mpegversion) {
+      case 1:
+        GST_DEBUG_OBJECT (pad, "Creating MPEG Audio, version 1 stream with " \
+            "PID 0x%04x", ts_data->pid);
+        ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_AUDIO_MPEG1, 
+            ts_data->pid);
+        break;
+      case 2:
+        GST_DEBUG_OBJECT (pad, "Creating MPEG Audio, version 2 stream with " \
+            "PID 0x%04x", ts_data->pid);
+        ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_AUDIO_MPEG2, 
+            ts_data->pid);
+        break;
+      case 4:
+      {
+        const GValue * value;
+        /* Codec data contains SPS/PPS which need to go in stream for valid ES */
+        value = gst_structure_get_value (s, "codec_data");
+        if (value) {
+          ts_data->codec_data = gst_buffer_ref (gst_value_get_buffer (value));
+          GST_DEBUG_OBJECT (pad, "we have additional codec data (%d bytes)",
+              GST_BUFFER_SIZE (ts_data->codec_data));
+          ts_data->prepare_func = flutsmux_prepare_aac;
+        }
+        else {
+          ts_data->codec_data = NULL;
+        }
+        GST_DEBUG_OBJECT (pad, "Creating MPEG Audio, version 4 stream with " \
+            "PID 0x%04x", ts_data->pid);
+        ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_AUDIO_AAC, 
+            ts_data->pid);
+        break;
+      }
+      default:
+        GST_WARNING_OBJECT (pad, "unsupported mpegversion %d", mpegversion);
+        goto beach;
     }
   }
   else if (gst_structure_has_name (s, "video/mpeg")) {
@@ -177,14 +289,14 @@ static GstFlowReturn flutsmux_create_stream (FluTsMux *mux,
     }
 
     if (mpegversion == 1) {
-      GST_DEBUG_OBJECT (mux, 
+      GST_DEBUG_OBJECT (pad, 
           "Creating MPEG Video, version 1 stream with PID 0x%04x", 
           ts_data->pid);
       ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_VIDEO_MPEG1, 
         ts_data->pid);
     }
     else {
-      GST_DEBUG_OBJECT (mux, 
+      GST_DEBUG_OBJECT (pad, 
           "Creating MPEG Video, version 2 stream with PID 0x%04x", 
           ts_data->pid);
       ts_data->stream = tsmux_create_stream (mux->tsmux, TSMUX_ST_VIDEO_MPEG2, 
@@ -247,6 +359,16 @@ flutsmux_choose_best_stream (FluTsMux *mux)
         ts_data->queued_buf = buf = gst_collect_pads_pop (mux->collect, c_data);
 
         if (buf != NULL) {
+          if (ts_data->prepare_func) {
+            buf = ts_data->prepare_func (buf, ts_data, mux);
+            if (buf) { /* Take the prepared buffer instead */
+              gst_buffer_unref (ts_data->queued_buf);
+              ts_data->queued_buf = buf;
+            }
+            else { /* If data preparation returned NULL, use unprepared one */
+              buf = ts_data->queued_buf;
+            }
+          }
           if (GST_BUFFER_TIMESTAMP (buf) != GST_CLOCK_TIME_NONE) {
             /* Ignore timestamps that go backward for now. FIXME: Handle all
              * incoming PTS */
@@ -301,7 +423,7 @@ static GstFlowReturn
 flutsmux_collected (GstCollectPads * pads, FluTsMux *mux)
 {
   GstFlowReturn ret = GST_FLOW_OK;
-  FluTsPadData *best;
+  FluTsPadData *best = NULL;
 
   GST_DEBUG_OBJECT (mux, "Pads collected");
 
@@ -313,8 +435,13 @@ flutsmux_collected (GstCollectPads * pads, FluTsMux *mux)
     best = flutsmux_choose_best_stream (mux);
 
     if (mux->pcr_stream == NULL) {
-      /* Take the first data stream for the PCR */
-      mux->pcr_stream = best->stream;
+      if (best) {
+        GstCollectData *c_data = (GstCollectData *) best;
+        /* Take the first data stream for the PCR */
+        GST_DEBUG_OBJECT (mux, "Use stream from pad %" GST_PTR_FORMAT " as PCR", 
+            c_data->pad);
+        mux->pcr_stream = best->stream;
+      }
     }
 
     /* Set the chosen PCR stream */
@@ -359,6 +486,9 @@ flutsmux_collected (GstCollectPads * pads, FluTsMux *mux)
         goto write_fail;
       }
     }
+    if (mux->pcr_stream == best->stream) {
+      mux->last_ts = best->last_ts;
+    }
   }
   else {
     /* FIXME: Drain all remaining streams */
@@ -402,6 +532,8 @@ static GstPad * flutsmux_request_new_pad (GstElement * element,
 
   pad_data->pid = pid;
   pad_data->last_ts = GST_CLOCK_TIME_NONE;
+  pad_data->codec_data = NULL;
+  pad_data->prepare_func = NULL;
 
   if (G_UNLIKELY (!gst_element_add_pad (element, pad)))
     goto could_not_add;
@@ -430,36 +562,126 @@ pad_failure:
 static void flutsmux_release_pad (GstElement * element, GstPad * pad)
 {
   FluTsMux *mux = FLU_TSMUX (element);
-
+  FluTsPadData *pad_data = NULL;
+  
   GST_DEBUG_OBJECT (mux, "Pad %" GST_PTR_FORMAT " being released", pad);
 
-  /* FIXME: Clean up data related to this stream */
-
+  /* Get the FluTsPadData out of the pad */
+  GST_OBJECT_LOCK (pad);
+  pad_data = (FluTsPadData *) gst_pad_get_element_private (pad);
+  if (G_LIKELY (pad_data)) {
+    /* Free codec data reference if any */
+    if (pad_data->codec_data) {
+      GST_DEBUG_OBJECT (element, "releasing codec_data reference");
+      gst_buffer_unref (pad_data->codec_data);
+      pad_data->codec_data = NULL;
+    }
+  }
+  GST_OBJECT_UNLOCK (pad);
+  
   gst_collect_pads_remove_pad (mux->collect, pad);
 }
 
-static gboolean new_packet_cb (guint8 *data, guint len, void *user_data)
+static gboolean 
+new_packet_cb (guint8 *data, guint len, void *user_data, gint64 new_pcr)
 {
   /* Called when the TsMux has prepared a packet for output. Return FALSE
    * on error */
   FluTsMux *mux = (FluTsMux *) user_data;
-  GstBuffer *buf;
+  GstBuffer *buf, *out_buf;
   GstFlowReturn ret;
+  gfloat current_ts;
+  gint64 m2ts_pcr, pcr_bytes, chunk_bytes;
+  gint8 *temp_ptr;
+  gint64 ts_rate;
 
-  GST_LOG_OBJECT (mux, "Outputting a packet of length %d", len);
-  ret = gst_pad_alloc_buffer (mux->srcpad, GST_BUFFER_OFFSET_NONE,
-            len, GST_PAD_CAPS (mux->srcpad), &buf);
-  if (G_UNLIKELY (ret != GST_FLOW_OK)) {
-    mux->last_flow_ret = ret;
-    return FALSE;
+  if (mux->m2ts_mode == TRUE){
+    /* Enters when the m2ts-mode is set true*/    
+    buf = gst_buffer_new_and_alloc (M2TS_PACKET_LENGTH);
+    if (G_UNLIKELY ( buf == NULL)) {
+      mux->last_flow_ret = GST_FLOW_ERROR;
+      return FALSE;
+    }
+    /* copies the ts data of 188 bytes to the m2ts buffer at an offset 
+       of 4 bytes of timestamp  */ 
+    memcpy (GST_BUFFER_DATA(buf) + 4, data, len);
+ 
+    if (new_pcr >= 0 ){
+      /*when there is a pcr value in ts data*/  
+      pcr_bytes = 0;
+      if (mux->first_pcr){
+        /*Incase of first pcr*/
+        /*writing the 4  byte timestamp value*/  
+        GST_WRITE_UINT32_BE (GST_BUFFER_DATA (buf), new_pcr);
+        
+        GST_LOG_OBJECT (mux, "Outputting a packet of length %d", 
+            M2TS_PACKET_LENGTH);
+        ret = gst_pad_push (mux->srcpad, buf);
+        if (G_UNLIKELY (ret != GST_FLOW_OK)) {
+          mux->last_flow_ret = ret;
+          return FALSE;
+        }
+        mux->first_pcr = FALSE;   
+        mux->previous_pcr = new_pcr;
+        pcr_bytes = M2TS_PACKET_LENGTH;
+      }
+      chunk_bytes = gst_adapter_available (mux->adapter);
+       
+      if (G_UNLIKELY (chunk_bytes)){
+        /* calculate rate based on latest and previous pcr values*/
+        ts_rate = ((chunk_bytes * STANDARD_TIME_CLOCK) / (new_pcr - 
+            mux->previous_pcr));
+        while (1){
+          /*loop till all the accumulated ts packets are transformed to 
+            m2ts packets and pushed*/
+          current_ts = ((gfloat)mux->previous_pcr/STANDARD_TIME_CLOCK) + 
+              ((gfloat)pcr_bytes / ts_rate );
+          m2ts_pcr = (((gint64)(STANDARD_TIME_CLOCK * current_ts / 300) & 
+              TWO_POW_33_MINUS1) * 300 ) + ((gint64)(STANDARD_TIME_CLOCK * 
+              current_ts) % 300 ); 
+          temp_ptr = (gint8 *)&m2ts_pcr; 
+ 
+          out_buf = gst_adapter_take_buffer (mux->adapter, 
+              M2TS_PACKET_LENGTH);
+          if (G_UNLIKELY (!out_buf))
+            break;
+ 
+          /*writing the 4  byte timestamp value*/
+          GST_WRITE_UINT32_BE (GST_BUFFER_DATA (out_buf), m2ts_pcr);
+          
+          GST_LOG_OBJECT (mux, "Outputting a packet of length %d", 
+              M2TS_PACKET_LENGTH);
+          ret = gst_pad_push (mux->srcpad, out_buf);
+          if (G_UNLIKELY (ret != GST_FLOW_OK)) {
+            mux->last_flow_ret = ret;
+            return FALSE;
+          }
+          pcr_bytes += M2TS_PACKET_LENGTH;
+        }
+        mux->previous_pcr = m2ts_pcr;
+      }
+    }
+    else
+    /* If theres no pcr in current ts packet then push the packet 
+       to an adapter, which is used to create m2ts packets*/
+      gst_adapter_push (mux->adapter, buf);
   }
+  else{
+    /* In case of Normal Ts packets*/
+    GST_LOG_OBJECT (mux, "Outputting a packet of length %d", len);
+    buf = gst_buffer_new_and_alloc (len);
+    if (G_UNLIKELY (buf == NULL)) {
+      mux->last_flow_ret = GST_FLOW_ERROR;
+      return FALSE;
+    }
 
-  memcpy (GST_BUFFER_DATA (buf), data, len);
-
-  ret = gst_pad_push (mux->srcpad, buf);
-  if (G_UNLIKELY (ret != GST_FLOW_OK)) {
-    mux->last_flow_ret = ret;
-    return FALSE;
+    memcpy (GST_BUFFER_DATA (buf), data, len);
+    GST_BUFFER_TIMESTAMP(buf) = mux->last_ts;
+    ret = gst_pad_push (mux->srcpad, buf);
+    if (G_UNLIKELY (ret != GST_FLOW_OK)) {
+      mux->last_flow_ret = ret;
+      return FALSE;
+    }
   }
 
   return TRUE;
@@ -470,8 +692,12 @@ flutsdemux_prepare_srcpad (FluTsMux *mux)
 {
   GstEvent *new_seg = 
       gst_event_new_new_segment (FALSE, 1.0, GST_FORMAT_BYTES, 0, -1, 0);
-  GstCaps * caps = 
-      gst_static_pad_template_get_caps (&flutsmux_src_factory);
+  GstCaps * caps = gst_caps_new_simple ("video/mpegts",
+         "systemstream", G_TYPE_BOOLEAN, TRUE,
+         "packetsize", G_TYPE_INT, (mux->m2ts_mode ? M2TS_PACKET_LENGTH : NORMAL_TS_PACKET_LENGTH),   
+         NULL);
+        
+//      gst_static_pad_template_get_caps (&flutsmux_src_factory);
 
   /* Set caps on src pad from our template and push new segment */
   gst_pad_set_caps (mux->srcpad, caps);
@@ -501,6 +727,10 @@ flutsmux_change_state (GstElement *element, GstStateChange transition)
     case GST_STATE_CHANGE_PAUSED_TO_READY:
       gst_collect_pads_stop (mux->collect);
       break;
+    case GST_STATE_CHANGE_READY_TO_NULL:
+      if (mux->adapter)
+        gst_adapter_clear (mux->adapter);
+      break;
     default:
       break;
   }
