@@ -1,13 +1,16 @@
-/* Copyright 2006 Fluendo S.L. 
- * Author: Jan Schmidt <jan@fluendo.com>
+/* Copyright 2006, 2007, 2008 Fluendo S.A. 
+ * Authors: Jan Schmidt <jan@fluendo.com>
+ *          Kapil Agrawal <kapil@fluendo.com>
+ *          Julien Moutte <julien@fluendo.com>
  * See the COPYING file in the top-level directory.
  */
-
+ 
 #ifndef __FLUTSMUX_H__
 #define __FLUTSMUX_H__
 
 #include <gst/gst.h>
 #include <gst/base/gstcollectpads.h>
+#include <gst/base/gstadapter.h>
 
 G_BEGIN_DECLS
 
@@ -20,6 +23,9 @@ typedef struct FluTsMux FluTsMux;
 typedef struct FluTsMuxClass FluTsMuxClass;
 typedef struct FluTsPadData FluTsPadData;
 
+typedef GstBuffer * (*FluTsPadDataPrepareFunction) (GstBuffer * buf,
+    FluTsPadData * data, FluTsMux * mux);
+
 struct FluTsMux {
   GstElement parent;
 
@@ -33,6 +39,12 @@ struct FluTsMux {
   gboolean first;
   TsMuxStream *pcr_stream;
   GstFlowReturn last_flow_ret;
+  GstAdapter *adapter;
+  gint64 previous_pcr;
+  gboolean m2ts_mode;
+  gboolean first_pcr;
+  
+  GstClockTime last_ts;
 };
 
 struct FluTsMuxClass  {
@@ -49,6 +61,10 @@ struct FluTsPadData {
   GstClockTime cur_ts; /* Adjusted TS for the pulled buffer */
   GstClockTime last_ts; /* Most recent valid TS for this stream */
 
+  GstBuffer * codec_data; /* Optional codec data available in the caps */
+
+  FluTsPadDataPrepareFunction prepare_func; /* Handler to prepare input data */
+
   gboolean eos;
 };
 
@@ -62,6 +78,11 @@ GType flutsmux_get_type (void);
 #define GSTTIME_TO_MPEGTIME(time) (gst_util_uint64_scale ((time), \
                         CLOCK_BASE, GST_MSECOND/10))
 
+#define NORMAL_TS_PACKET_LENGTH 188
+#define M2TS_PACKET_LENGTH      192
+#define STANDARD_TIME_CLOCK     27000000
+/*33 bits as 1 ie 0x1ffffffff*/
+#define TWO_POW_33_MINUS1     ((0xffffffff * 2) - 1) 
 G_END_DECLS
 
 #endif
